@@ -1,3 +1,7 @@
+/*
+ * SPDX-License-Identifier: Apache-2.0
+ */
+
 // Adapter for Dropout in default domain from version 6 to 7
 
 #pragma once
@@ -17,8 +21,9 @@ class Dropout_6_7 final : public Adapter {
       }
     }
 
-    void adapt(std::shared_ptr<Graph> graph, Node* node) const override {
+    Node* adapt(std::shared_ptr<Graph> graph, Node* node) const override {
       adapt_dropout_6_7(graph, node);
+      return node;
     }
 };
 
